﻿using System;
using System.Collections.Generic;
using System.Device.Location;
using System.IO.IsolatedStorage;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Navigation;

using Microsoft.Devices.Sensors;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input.Touch;

using OrientingMap.ImageryService;

namespace OrientingMap
{
    public partial class MainPage : PhoneApplicationPage
    {
        const int BITRES = 29;

        GeoCoordinateWatcher geoWatcher = new GeoCoordinateWatcher(GeoPositionAccuracy.High);
        Motion motion;

        string roadUriTemplate, aerialUriTemplate;
        MapStyle mapStyle;
        int minimumLevel, maximumLevel;

        double centerRelativeLongitude;
        double centerRelativeLatitude;
        double locationRelativeLongitude;
        double locationRelativeLatitude;

        int baseLevel;
        double canvasScale = 1;
        double zoomFactor;
        Dictionary<QuadKey, MapTile> currentMapTiles = new Dictionary<QuadKey, MapTile>();
        List<QuadKey> validQuadKeys = new List<QuadKey>();
        List<QuadKey> deleteList = new List<QuadKey>();
        Vector2 canvasOrigin;
        double accumulatedDeltaX, accumulatedDeltaY;

        public MainPage()
        {
            InitializeComponent();

            // Fix that annoying ApplicationBar bug
            gotoMeAppBarButton = this.ApplicationBar.Buttons[1] as ApplicationBarIconButton;

            // Enable XNA TouchPanel gestures
            TouchPanel.EnabledGestures = GestureType.FreeDrag | GestureType.DragComplete |
                                         GestureType.Pinch | GestureType.PinchComplete;

            // Initialize the Motion sensor
            if (Motion.IsSupported)
            {
                motion = new Motion();
                motion.TimeBetweenUpdates = TimeSpan.FromMilliseconds(33);
                motion.CurrentValueChanged += OnMotionCurrentValueChanged;
            }

            // Two more event handlers
            geoWatcher.PositionChanged += OnGeoWatcherPositionChanged;
            Loaded += OnMainPageLoaded;
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            // Get saved information
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;

            // Initialize to 2010 mean center of US population
            if (!settings.TryGetValue<double>("Longitude", out centerRelativeLongitude))
                centerRelativeLongitude = GetRelativeLongitude(-90.18471);

            if (!settings.TryGetValue<double>("Latitude", out centerRelativeLatitude))
                centerRelativeLatitude = GetRelativeLatitude(38.62452);

            if (!settings.TryGetValue<double>("ZoomFactor", out zoomFactor))
                zoomFactor = 12;

            if (!settings.TryGetValue<MapStyle>("MapStyle", out mapStyle))
                mapStyle = MapStyle.Road;

            // Start the Motion sensor and GPS watcher
            if (motion != null)
            {
                try { motion.Start(); }
                catch { }
            }

            geoWatcher.Start();
            base.OnNavigatedTo(args);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs args)
        {
            // Stop the Motion sensor and GPS watcher
            if (motion != null)
                motion.Stop();

            geoWatcher.Stop();

            // Save information
            IsolatedStorageSettings settings = IsolatedStorageSettings.ApplicationSettings;
            settings["Longitude"] = centerRelativeLongitude;
            settings["Latitude"] = centerRelativeLatitude;
            settings["ZoomFactor"] = zoomFactor;
            settings["MapStyle"] = mapStyle;
            settings.Save();

            base.OnNavigatedFrom(args);
        }

        void OnMainPageLoaded(object sender, RoutedEventArgs e)
        {
            // Get the Canvas (0,0) origin for touch processing
            System.Windows.Point ptOrigin = imageCanvas.TransformToVisual(null).Transform(new System.Windows.Point(0, 0));
            canvasOrigin = new Vector2((float)ptOrigin.X, (float)ptOrigin.Y);

            // Initialize the Bing Maps imagery service
            ImageryServiceClient imageryServiceClient = new ImageryServiceClient("BasicHttpBinding_IImageryService");
            imageryServiceClient.GetImageryMetadataCompleted += GetImageryMetadataCompleted;

            // Make a request for the road metadata
            ImageryMetadataRequest request = new ImageryMetadataRequest
            {
                Credentials = new Credentials
                {
                    ApplicationId = "put your own Bing credentials GUID here"
                },
                Style = MapStyle.Road
            };
            imageryServiceClient.GetImageryMetadataAsync(request, "road");

            // Make a request for the aerial metadata
            request.Style = MapStyle.Aerial;
            imageryServiceClient.GetImageryMetadataAsync(request, "aerial");
        }

        void GetImageryMetadataCompleted(object sender, GetImageryMetadataCompletedEventArgs args)
        {
            if (!args.Cancelled && args.Error == null)
            {
                // Get the "powered by" bitmap
                poweredByBitmap.UriSource = args.Result.BrandLogoUri;
                poweredByDisplay.Visibility = Visibility.Visible;

                // Get the range of map levels available
                ImageryMetadataResult result = args.Result.Results[0];
                minimumLevel = result.ZoomRange.From;
                maximumLevel = result.ZoomRange.To;

                // Get the URI and make some substitutions
                string uri = result.ImageUri;
                uri = uri.Replace("{subdomain}", result.ImageUriSubdomains[0]);
                uri = uri.Replace("&token={token}", "");
                uri = uri.Replace("{culture}", "en-US");

                if (args.UserState as string == "road")
                    roadUriTemplate = uri;
                else
                    aerialUriTemplate = uri;

                // If both URIs are available, let's begin
                if (roadUriTemplate != null && aerialUriTemplate != null)
                {
                    QuadKey.UriTemplate = mapStyle == MapStyle.Road ? roadUriTemplate : aerialUriTemplate;

                    // Calculate baselevel and canvasScale from zoomFactor
                    baseLevel = (int)zoomFactor;
                    canvasScale = Math.Pow(2, zoomFactor - baseLevel);
                    imageCanvasScale.ScaleX = canvasScale;
                    imageCanvasScale.ScaleY = canvasScale;

                    // Create the initial tiles
                    RefreshDisplay();
                }

                // Use CompositionTarget.Rendering for touch processing
                CompositionTarget.Rendering += OnCompositionTargetRendering;
            }
            else
            {
                errorTextBlock.Text = "Cannot access Bing Maps: " + args.Error.Message;
            }
        }

        void OnMotionCurrentValueChanged(object sender, SensorReadingEventArgs<MotionReading> args)
        {
            this.Dispatcher.BeginInvoke(() =>
            {
                // Get the rotation matrix & convert to horizontal coordinates
                Microsoft.Xna.Framework.Matrix matrix = args.SensorReading.Attitude.RotationMatrix;
                Vector3 vector = Vector3.Transform(Vector3.UnitY, matrix);

                // Set the azimuth angle to rotation transforms
                double azimuth = 180 * Math.Atan2(vector.X, vector.Y) / Math.PI;
                imageCanvasRotate.Angle = azimuth;
                northArrowRotate.Angle = azimuth;
            });
        }

        void OnGeoWatcherPositionChanged(object sender,
                                 GeoPositionChangedEventArgs<GeoCoordinate> args)
        {
            // Save the longitude and latitude, set set the location on the display
            locationRelativeLongitude = GetRelativeLongitude(args.Position.Location.Longitude);
            locationRelativeLatitude = GetRelativeLatitude(args.Position.Location.Latitude);
            SetLocationPosition();

            // For the first time this event is fired:
            // Make the location circle visible
            locationDisplay.Visibility = Visibility.Visible;

            // And enable the app bar button to move to location
            gotoMeAppBarButton.IsEnabled = true;
        }

        void OnToggleViewAppBarButtonClick(object sender, EventArgs args)
        {
            // Toggle the map style
            if (mapStyle == MapStyle.Road)
                mapStyle = MapStyle.Aerial;
            else
                mapStyle = MapStyle.Road;

            // Set the URI template
            QuadKey.UriTemplate = mapStyle == MapStyle.Road ? roadUriTemplate : aerialUriTemplate;

            // Refresh the tiles with the new map style
            foreach (QuadKey quadKey in currentMapTiles.Keys)
                currentMapTiles[quadKey].Refresh();
        }

        void OnGotoMeAppBarButtonClick(object sender, EventArgs args)
        {
            centerRelativeLongitude = locationRelativeLongitude;
            centerRelativeLatitude = locationRelativeLatitude;

            imageCanvas.Children.Clear();
            currentMapTiles.Clear();
            RefreshDisplay();
        }

        // Touch processing using XNA TouchPanel
        void OnCompositionTargetRendering(object sender, EventArgs args)
        {
            while (TouchPanel.IsGestureAvailable)
            {
                GestureSample gesture = TouchPanel.ReadGesture();

                switch (gesture.GestureType)
                {
                    case GestureType.FreeDrag:
                        // Adjust delta for rotation of canvas
                        Vector2 delta = TransformGestureToMap(gesture.Delta);

                        // Translate the canvas
                        imageCanvasTranslate.X += delta.X;
                        imageCanvasTranslate.Y += delta.Y;

                        // Adjust the center longitude and latitude
                        centerRelativeLongitude -= delta.X / (1 << baseLevel + 8) / canvasScale;
                        centerRelativeLatitude -= delta.Y / (1 << baseLevel + 8) / canvasScale;

                        // Accumulate the panning distance
                        accumulatedDeltaX += delta.X;
                        accumulatedDeltaY += delta.Y;

                        // Check if that's sufficient to warrant a screen refresh
                        if (Math.Abs(accumulatedDeltaX) > 256 ||
                            Math.Abs(accumulatedDeltaY) > 256)
                        {
                            RefreshDisplay();
                            accumulatedDeltaX = 0;
                            accumulatedDeltaY = 0;
                        }
                        break;

                    case GestureType.DragComplete:
                        Cleanup();
                        break;

                    case GestureType.Pinch:
                        // Get the old and new finger positions relative to canvas origin
                        Vector2 newPoint1 = gesture.Position - canvasOrigin;
                        Vector2 oldPoint1 = newPoint1 - gesture.Delta;

                        Vector2 newPoint2 = gesture.Position2 - canvasOrigin;
                        Vector2 oldPoint2 = newPoint2 - gesture.Delta2;

                        // Rotate in accordance with the current rotation angle
                        oldPoint1 = TransformGestureToMap(oldPoint1);
                        newPoint1 = TransformGestureToMap(newPoint1);
                        oldPoint2 = TransformGestureToMap(oldPoint2);
                        newPoint2 = TransformGestureToMap(newPoint2);

                        // Find the delta scaling factor
                        double deltaScale = (newPoint2 - newPoint1).Length() / (oldPoint2 - oldPoint1).Length();

                        // But limit it so zoomFactor ranges from 1 to 21
                        double minDeltaScale = Math.Pow(2, minimumLevel - baseLevel) / canvasScale;
                        double maxDeltaScale = Math.Pow(2, maximumLevel - baseLevel) / canvasScale;
                        deltaScale = Math.Min(maxDeltaScale, Math.Max(minDeltaScale, deltaScale));

                        // Find oldScale and newScale
                        double oldScale = canvasScale;
                        double newScale = oldScale * deltaScale;

                        // Use that to adjust the Canvas scale
                        canvasScale *= deltaScale;
                        imageCanvasScale.ScaleX = canvasScale;
                        imageCanvasScale.ScaleY = canvasScale;

                        // Calculate the new zoomFactor
                        zoomFactor = baseLevel + Math.Log(canvasScale) / Math.Log(2);

                        // Now find the translation factor implied by the gesture deltas
                        Vector2 oldTranslate = new Vector2((float)imageCanvasTranslate.X, (float)imageCanvasTranslate.Y);
                        Vector2 translate1 = newPoint1 - (float)deltaScale * (oldPoint1 - oldTranslate);
                        Vector2 translate2 = newPoint2 - (float)deltaScale * (oldPoint2 - oldTranslate);

                        // Take the average and use that for the new translation factors
                        Vector2 newTranslate = 0.5f * (translate1 + translate2);
                        imageCanvasTranslate.X = newTranslate.X;
                        imageCanvasTranslate.Y = newTranslate.Y;

                        // Adjust the center longitude and latitude similarly
                        centerRelativeLongitude -= newTranslate.X / (1 << baseLevel + 8) / newScale - oldTranslate.X / (1 << baseLevel + 8) / oldScale;
                        centerRelativeLatitude -= newTranslate.Y / (1 << baseLevel + 8) / newScale - oldTranslate.Y / (1 << baseLevel + 8) / oldScale;

                        // If the Canvas has doubled in size, shift the scaling to the tiles
                        while (canvasScale >= 2)
                        {
                            baseLevel += 1;
                            canvasScale /= 2;
                            imageCanvasScale.ScaleX = canvasScale;
                            imageCanvasScale.ScaleY = canvasScale;

                            foreach (UIElement child in imageCanvas.Children)
                            {
                                (child as MapTile).Scale *= 2;
                                Canvas.SetLeft(child, 2 * Canvas.GetLeft(child));
                                Canvas.SetTop(child, 2 * Canvas.GetTop(child));
                            }
                        }

                        // Also if the Canvas has halved its size
                        while (canvasScale <= 0.5)
                        {
                            baseLevel -= 1;
                            canvasScale *= 2;
                            imageCanvasScale.ScaleX = canvasScale;
                            imageCanvasScale.ScaleY = canvasScale;

                            foreach (UIElement child in imageCanvas.Children)
                            {
                                (child as MapTile).Scale /= 2;
                                Canvas.SetLeft(child, Canvas.GetLeft(child) / 2);
                                Canvas.SetTop(child, Canvas.GetTop(child) / 2);
                            }
                        }

                        RefreshDisplay();
                        break;

                    case GestureType.PinchComplete:
                        Cleanup();
                        break;
                }
            }
        }

        Vector2 TransformGestureToMap(Vector2 vector)
        {
            double radians = Math.PI * imageCanvasRotate.Angle / 180;
            double x = vector.X * Math.Cos(radians) + vector.Y * Math.Sin(radians);
            double y = -vector.X * Math.Sin(radians) + vector.Y * Math.Cos(radians);
            return new Vector2((float)x, (float)y);
        }

        void RefreshDisplay()
        {
            validQuadKeys.Clear();

            // Find integer longitude and latitude
            int centerLongitude = (int)(centerRelativeLongitude * (1 << BITRES));
            int centerLatitude = (int)(centerRelativeLatitude * (1 << BITRES));

            // Crop them based on the base level, and get QuadKey
            int longitude = centerLongitude >> BITRES - baseLevel;
            int latitude = centerLatitude >> BITRES - baseLevel;
            QuadKey baseQuadKey = new QuadKey(longitude, latitude, baseLevel);

            // Calculate offsets based on location within tile
            double xOffset = (centerLongitude >> BITRES - baseLevel - 8) % 256;
            xOffset = (centerLongitude / (double)(1 << BITRES - baseLevel - 8)) % 256;

            double yOffset = (centerLatitude >> BITRES - baseLevel - 8) % 256;
            yOffset = (centerLatitude / (double)(1 << BITRES - baseLevel - 8)) % 256;

            xOffset += imageCanvasTranslate.X / canvasScale;
            yOffset += imageCanvasTranslate.Y / canvasScale;

            // Number of tiles to create depends on current canvasScale
            int scope = (int)Math.Ceiling(3 / canvasScale);

            for (int row = -scope; row <= scope; row++)
                for (int col = -scope; col <= scope; col++)
                {
                    QuadKey quadKey = baseQuadKey.GetQuadKeyFromRowColumnOffsets(col, row);

                    if (quadKey == null)
                        continue;

                    validQuadKeys.Add(quadKey);

                    if (!currentMapTiles.ContainsKey(quadKey))
                    {
                        double x = 256 * col - xOffset;
                        double y = 256 * row - yOffset;

                        x = Math.Round(x);
                        y = Math.Round(y);

                        // Create MapTile
                        MapTile mapTile = new MapTile
                        {
                            ImageOpened = OnMapTileImageOpened,
                            QuadKey = quadKey
                        };

                        Canvas.SetLeft(mapTile, x);
                        Canvas.SetTop(mapTile, y);
                        Canvas.SetZIndex(mapTile, quadKey.Level);
                        imageCanvas.Children.Add(mapTile);
                        currentMapTiles.Add(quadKey, mapTile);
                    }
                }

            SetLocationPosition();
        }

        void Cleanup()
        {
            deleteList.Clear();

            foreach (QuadKey quadKey in currentMapTiles.Keys)
            {
                if (quadKey.Level == baseLevel)
                {
                    if (!validQuadKeys.Contains(quadKey))
                        deleteList.Add(quadKey);
                }
                else if (quadKey.Level < baseLevel)
                {
                    bool tileStillVisible = false;

                    foreach (QuadKey childQuadKey in quadKey.Children)
                        if (currentMapTiles.ContainsKey(childQuadKey) &&
                            !currentMapTiles[childQuadKey].IsImageOpened)
                        {
                            tileStillVisible = true;
                            break;
                        }

                    if (!tileStillVisible)
                        deleteList.Add(quadKey);
                }
                else
                {
                    QuadKey parentQuadKey = quadKey.Parent;

                    if (!currentMapTiles.ContainsKey(parentQuadKey) || currentMapTiles[parentQuadKey].IsImageOpened)
                        deleteList.Add(quadKey);
                }
            }

            foreach (QuadKey quadKey in deleteList)
            {
                MapTile mapTile = currentMapTiles[quadKey];
                mapTile.QuadKey = null;
                mapTile.ImageOpened = null;

                imageCanvas.Children.Remove(mapTile);
                currentMapTiles.Remove(quadKey);
            }
        }

        void OnMapTileImageOpened(object sender, EventArgs args)
        {
            Cleanup();
        }

        // Also called from geo watcher
        void SetLocationPosition()
        {
            locationTranslate.X = 256 * (1 << baseLevel) * (locationRelativeLongitude - centerRelativeLongitude) * imageCanvasScale.ScaleX - imageCanvasTranslate.X;
            locationTranslate.Y = 256 * (1 << baseLevel) * (locationRelativeLatitude - centerRelativeLatitude) * imageCanvasScale.ScaleY - imageCanvasTranslate.Y;
        }

        double GetRelativeLongitude(double longitude)
        {
            return (180 + longitude) / 360;
        }

        double GetRelativeLatitude(double latitude)
        {
            double sin = Math.Sin(Math.PI * latitude / 180);
            return 0.5 - Math.Log((1 + sin) / (1 - sin)) / (4 * Math.PI);
        }
    }
}